import dataclasses
import logging
import os
import shutil
import subprocess
import sys
import tempfile
import urllib.parse

import git
import yaml

import gci.componentmodel as cm

import ci.util
import concourse.enumerator as ce

logger = logging.getLogger('pipeline-cli')

own_dir = os.path.abspath(os.path.dirname(__file__))


def _branch_cfg(
    repo: git.Repo,
    meta_ci_ref: str,
    fetch: bool=False,
    absent_ok: bool=True,
) -> ce.BranchCfg | None:
    '''
    parses `branch.cfg` from repo, assuming it is a valid yaml-file named `branch.cfg`
    '''
    if fetch:
        remote = repo.remote()
        remote.fetch(f'refs/meta/ci:{meta_ci_ref}')

    for ref in repo.refs:
        if ref.path == meta_ci_ref:
            break
    else:
        if absent_ok:
            return None
        raise RuntimeError(f'{meta_ci_ref=} not available in local {repo=}')

    for blob in ref.object.tree.blobs:
        if blob.name == 'branch.cfg':
            break
    else:
        raise RuntimeError(f'did not find regular file `branch.cfg` in {meta_ci_ref}')

    branch_cfg = yaml.safe_load(blob.data_stream)
    return ce.BranchCfg(raw_dict=branch_cfg)


def _pipeline_definitions(
    repo: git.Repo,
    branch_name: str,
    branch_cfg: ce.BranchCfg | None=None,
):
    '''
    parses pipeline-definitions from .ci/pipeline_definitions from current repo's work-tree
    '''
    with open(os.path.join(repo.working_tree_dir, '.ci', 'pipeline_definitions')) as f:
        raw = yaml.safe_load(f)

    if not branch_cfg:
        return raw

    cfg_entry = branch_cfg.cfg_entry_for_branch(branch=branch_name)

    if not cfg_entry:
        return raw

    return ci.util.merge_dicts(
        raw,
        cfg_entry.override_definitions(),
    )


def _component_name(
    repo: git.Repo,
    job: dict,
):
    try:
        component_descriptor_trait = job['traits']['component_descriptor']
        if name := component_descriptor_trait['component_name']:
            return name
    except (KeyError, TypeError):
        pass

    remote = repo.remote()
    logger.info(f'{remote.name=}')

    url = urllib.parse.urlparse(next(remote.urls))
    hostname = url.hostname.removeprefix('gardener.')
    component_name = f'{hostname}{url.path}'
    return component_name


def _main_source(
    repo: git.Repo,
    version: str,
    job: dict,
):
    return cm.ComponentSource(
        name='dummy-name',
        version=version,
        access=cm.GithubAccess(
            type=cm.AccessType.GITHUB,
            repoUrl=next(repo.remote().urls),
            ref=repo.active_branch.name,
            commit=repo.head.commit.hexsha,
        ),
        labels=[
            cm.Label(
                name='cloud.gardener/cicd/source',
                value={
                    'repository-classification': 'main',
                },
            ),
        ],
    )


def _iter_resources(
    version: str,
    job: dict,
):
    if not (traits := job.get('traits', None)):
        return

    if not (publish_trait := traits.get('publish', None)):
        return

    if not (images := publish_trait.get('dockerimages', None)):
        return

    for name, image in images.items():
        if 'image' in image:
            yield cm.Resource(
                name=name,
                version=version,
                type=cm.ArtefactType.OCI_IMAGE,
                access=cm.OciAccess(
                    type=cm.AccessType.OCI_REGISTRY,
                    imageReference=image['image'] + ':' + version,
                ),
                labels=image.get('resource_labels', []),
            )
        else:
            for target in image['targets']:
                yield cm.Resource(
                    name=target['name'],
                    version=version,
                    type=cm.ArtefactType.OCI_IMAGE,
                    access=cm.OciAccess(
                        type=cm.AccessType.OCI_REGISTRY,
                        imageReference=target['image'] + ':' + version,
                    ),
                    labels=image.get('resource_labels', []),
                )


def _repo(repo: str=None):
    if isinstance(repo, git.Repo):
        return repo

    if not repo:
        repo = os.getcwd()

    try:
        repo = git.Repo(
            path=repo,
            search_parent_directories=True,
        )
        return repo
    except git.exc.InvalidGitRepositoryError: # pylint: disable=E1101
        logger.error(f'not a git-repository: {repo}. Hint: change PWD, or pass --repo')
        exit(1)


def base_component_descriptor(
    repo: str=None,
    meta_ci: str='if-local', # | fetch
    meta_ci_ref: str='refs/meta/ci',
    pipeline_name: str=None, # only required if there is more than one
    job_name: str=None,
    component_name: str=None,
    ocm_repo: str='europe-docker.pkg.dev/gardener-project/public',
    version: str=None,
    outfile: str=None,
):
    repo = _repo(repo=repo)

    try:
        branch_name = repo.active_branch.name
    except TypeError:
        logger.error('repository must not be in detched-head state')
        exit(1)

    logger.info(f'using {branch_name=}')
    logger.info('looking for branch-cfg (refs/meta/ci)')
    branch_cfg = _branch_cfg(
        repo=repo,
        meta_ci_ref=meta_ci_ref,
        fetch=meta_ci == 'fetch',
        absent_ok=meta_ci == 'if-local',
    )
    if not branch_cfg:
        logger.info('did not find branch-cfg (ignoring)')

    pipeline_definitions = _pipeline_definitions(
        repo=repo,
        branch_name=branch_name,
        branch_cfg=branch_cfg,
    )

    if not pipeline_definitions:
        logger.error('pipeline-definitions appear to be empty')
        exit(1)

    if not pipeline_name and len(pipeline_definitions) > 1:
        logger.info('more than one pipeline found - will guess')

    if not pipeline_name:
        pipeline_name, pipeline_definition = next(pipeline_definitions.items().__iter__())
    else:
        pipeline_definition = pipeline_definitions[pipeline_name]

    logger.info(f'{pipeline_name=}')

    best_score = 0
    best_candidate = None
    best_job_name = None

    if not job_name:
        logger.info('guessing best job-name, as none was specified')

    for p_name, pipeline_definition in pipeline_definitions.items():
        if pipeline_name and pipeline_name != p_name:
            continue

        if not (jobs := pipeline_definition.get('jobs', None)):
            logger.error(f'local {pipeline_name=} contained no jobs')
            logger.info('hint:fetch refs/meta/ci (--meta-ci=fetch)')
            exit(1)

        base_definition = pipeline_definition.get('base_definition', {})

        for j_name, job in jobs.items():
            jobs[j_name] = ci.util.merge_dicts(base_definition, job)

        for j_name, job in jobs.items():
            if job_name and job_name != j_name:
                continue

            traits = job.get('traits', {})
            score = 0
            if not 'component_descriptor' in traits:
                continue # not a candidate

            # jobs w/ release-trait publish "final" component-descriptors, thus they should
            # be strongly preferred over jobs that do not have release-trait
            if 'release' in traits:
                score += 2

            # publish-trait contribute to base-component-descriptor
            if 'publish' in traits:
                score += 1

            if score <= best_score:
                continue # candidate was not better

            best_candidate = job
            best_score = score
            best_job_name = j_name

            logger.info('guessed job-name (pass explicitly if guessed wrong)')
            job = best_candidate

    job_name = best_job_name
    job = best_candidate

    if not job_name:
        logger.error('did not find any job w/ at least component_descriptor trait')
        logger.info('hint: fetch refs/meta/ci (--meta-ci=fetch)')
        exit(1)

    logger.info(f'{job_name=}')

    if not component_name:
        component_name = _component_name(
            repo=repo,
            job=job,
        )
        logger.info(f'guessed {component_name=} (from default remote-url)')

    if not version:
        if os.path.isfile(versionfile := os.path.join(repo.working_tree_dir, 'VERSION')):
            version = open(versionfile).read().strip()
            if version.startswith('v'):
                prefix = 'v'
            else:
                prefix = ''
            import version as version_mod
            version = version_mod.parse_to_semver(version)
            version = str(version.finalize_version())
        else:
            version = f'{prefix}1.2.3'

    component_descriptor = cm.ComponentDescriptor(
        meta=cm.Metadata(),
        component=cm.Component(
            name=component_name,
            version=version,
            repositoryContexts=[
                cm.OciOcmRepository(
                    type=cm.AccessType.OCI_REGISTRY,
                    baseUrl=ocm_repo,
                    subPath='',
                ),
            ],
            provider='sap-se',
            componentReferences=[],
            resources=[
                r for r in _iter_resources(
                    version=version,
                    job=job,
                )
            ],
            sources=[
                _main_source(
                    repo=repo,
                    job=job,
                    version=version,
                ),
            ],
        ),
    )

    if outfile:
        outfileh = open(outfile, 'w')
    else:
        outfileh = sys.stdout

    yaml.dump(
        data=dataclasses.asdict(component_descriptor),
        stream=outfileh,
        Dumper=cm.EnumValueYamlDumper,
    )

    return component_descriptor


def component_descriptor(
    repo: str=None,
    meta_ci: str='if-local', # | fetch
    meta_ci_ref: str='refs/meta/ci',
    pipeline_name: str=None, # only required if there is more than one
    job_name: str=None,
    component_name: str=None,
    version: str=None,
    outfile: str=None,
    component_repo: str='europe-docker.pkg.dev/gardener-project/public',
    component_descriptor_script: str=None,
    base_component_descriptor_path: str=None,
):
    repo = _repo(repo=repo)
    if not outfile:
        outfile = os.path.join(os.getcwd(), 'component-descriptor.yaml')

    if not component_descriptor_script:
        component_descriptor_script = os.path.join(
            repo.working_tree_dir,
            '.ci',
            'component_descriptor',
        )

    base_component_descriptor_file = tempfile.NamedTemporaryFile()
    logger.info(f'{base_component_descriptor_file.name=}')
    logger.info(f'{outfile=}')

    if not base_component_descriptor_path:
        base_descriptor  = base_component_descriptor(
            repo=repo,
            meta_ci=meta_ci,
            meta_ci_ref=meta_ci_ref,
            pipeline_name=pipeline_name,
            job_name=job_name,
            component_name=component_name,
            version=version,
            outfile=base_component_descriptor_file.name,
        )
    else:
        shutil.copyfile(base_component_descriptor_path, base_component_descriptor_file.name)
        with open(base_component_descriptor_path) as f:
            raw = yaml.safe_load(f)
        base_descriptor = cm.ComponentDescriptor.from_dict(raw)

    base_component = base_descriptor.component

    if not os.path.isfile(component_descriptor_script):
        logger.info(f'{component_descriptor_script=} is not a file - skipping callback')

        shutil.copyfile(
            base_component_descriptor_file.name,
            outfile,
        )
        logger.info(f'copied to {outfile=}')
        exit(0)

    cli = os.path.join(own_dir, 'cli_gen.py')
    dependencies_cmd = ' '.join((
        cli,
        'productutil_v2',
        'add_dependencies',
        '--descriptor-src-file', base_component_descriptor_file.name,
        '--descriptor-out-file', base_component_descriptor_file.name,
        '--component-version', base_component.version,
        '--component-name', base_component.name,
    ))

    env = os.environ.copy()
    env |= {
        'MAIN_REPO_DIR': repo.working_tree_dir,
        'BASE_DEFINITION_PATH': base_component_descriptor_file.name,
        'COMPONENT_DESCRIPTOR_PATH': outfile,
        'COMPONENT_NAME': base_component.name,
        'COMPONENT_VERSION': base_component.version,
        'EFFECTIVE_VERSION': base_component.version,
        'CURRENT_COMPONENT_REPOSITORY': component_repo,
        'ADD_DEPENDENCIES_CMD': dependencies_cmd,
    }

    subprocess.run(
        (component_descriptor_script,),
        env=env,
        check=True,
    )

    logger.info(f'component-descriptor should be at: {outfile=}')



'''
temporary (hopefully short..) migration util as a backwards-compatible replacement for
productutil.py -> add_dependencies

different to the aforementioned cmd, a component-descriptor-v2 is expected as input, and
created as output
'''

import dataclasses
import sys
import yaml

import gci.componentmodel as cm

import ci.util

CliHint = ci.util.CliHint

parse = yaml.safe_load


def _raw_component_dep_to_v2(raw: dict):
  if not 'componentName' in raw:
    component_name = raw['name'].strip()
    name = component_name.replace('/', '_').replace('.', '_')
  else:
    component_name = raw['componentName'].strip()
    name = raw['name'].strip()

  version = raw['version'].strip()

  args = {
    'componentName': component_name,
    'name': name,
    'version': version,
  }

  if 'labels' in raw:
    args['labels'] = raw['labels']

  return cm.ComponentReference(**args)


def _raw_image_dep_to_v2(raw: dict):
  img_ref = raw['image_reference']
  args = {
    'name': raw['name'],
    'version': raw['version'],
    'type': cm.ResourceType.OCI_IMAGE,
    'relation': cm.ResourceRelation(raw.get('relation', cm.ResourceRelation.EXTERNAL)),
    'access': cm.OciAccess(type=cm.AccessType.OCI_REGISTRY, imageReference=img_ref),
  }

  if 'labels' in raw:
    args['labels'] = raw['labels']

  return cm.Resource(**args)


def _raw_generic_dep_to_v2(raw: dict):
  name = raw['name']
  version = raw['version']
  rel = cm.ResourceRelation(raw.get('relation', cm.ResourceRelation.LOCAL))

  return cm.Resource(
    name=name,
    version=version,
    type='generic', # removed in later OCM versions, kept for backwards compatibility
    relation=rel,
    access=None,
  )


def add_dependencies(
  descriptor_src_file: str,
  component_name: str,
  component_version: str,
  descriptor_out_file: str=None,
  component_dependencies: [str]=[],
  container_image_dependencies: [str]=[],
  generic_dependencies: [str]=[],
):
  component_descriptor = cm.ComponentDescriptor.from_dict(
    ci.util.parse_yaml_file(descriptor_src_file)
  )
  component = component_descriptor.component

  # perform sanity checks (component_name and version *must* match)
  if not component.name == component_name:
    raise ValueError(f'{component_name=} != {component.name=}')
  if not component.version == component_version:
    raise ValueError(f'{component_version=} != {component.version=}')

  component.componentReferences += [
    _raw_component_dep_to_v2(parse(cdep)) for cdep in component_dependencies
  ]
  for img_dep in container_image_dependencies:
    img_dep = parse(img_dep)
    img_res = _raw_image_dep_to_v2(img_dep)

    component.resources.append(img_res)

  for gen_dep in generic_dependencies:
    gen_dep = parse(gen_dep)
    gen_res = _raw_generic_dep_to_v2(gen_dep)
    component.resources.append(gen_res)

  if descriptor_out_file:
    outfh = open(descriptor_out_file, 'w')
  else:
    outfh = sys.stdout

  yaml.dump(
    data=dataclasses.asdict(component_descriptor),
    Dumper=cm.EnumValueYamlDumper,
    stream=outfh,
  )



import concurrent.futures
import enum
import hashlib
import json
import pprint
import subprocess
import sys
import tabulate
import tarfile
import tempfile

import requests

import ccc.oci
import ctx
import delivery.client
import delivery.util
import oci
import oci.model as om
import oci.workarounds as ow
import tarutil
import unixutil.scan as us
import version

__cmd_name__ = 'oci'


class OciManifestChoice(enum.Enum):
    PREFER_MULTIARCH = 'prefer-multiarch'
    SINGLE = 'single'
    MULTIARCH = 'multiarch'


def cp(
    src:str,
    tgt:str,
    annotations:[str]=list(),
):
    if annotations:
        annotations_dict = {}
        for a in annotations:
            k, v = a.split('=')
            annotations_dict[k] = v
    else:
        annotations_dict = None

    oci_client = ccc.oci.oci_client()

    oci.replicate_artifact(
        src_image_reference=src,
        tgt_image_reference=tgt,
        oci_client=oci_client,
        mode=oci.ReplicationMode.PREFER_MULTIARCH,
        annotations=annotations_dict,
    )


def tags(image: str):
    oci_client = ccc.oci.oci_client()

    print('\n'.join(oci_client.tags(image_reference=image)))


def _manifest(
    image_reference: str,
    oci_client,
):
    image_reference: om.OciImageReference = om.OciImageReference.to_image_ref(image_reference)
    manifest = oci_client.manifest(
        image_reference=image_reference,
        accept=om.MimeTypes.prefer_multiarch,
    )

    if isinstance(manifest, om.OciImageManifestList):
        manifest: om.OciImageManifestList
        manifest: om.OciImageManifestListEntry = manifest.manifests[0]
        sub_img_ref = f'{image_reference.ref_without_tag}@{manifest.digest}'
        manifest = oci_client.manifest(sub_img_ref)

    return manifest


def ls(
    image: str,
    long: bool=False, # inspired by ls --long ; aka: ls -l
):
    oci_client = ccc.oci.oci_client()
    image_reference: om.OciImageReference = om.OciImageReference.to_image_ref(image)

    manifest = _manifest(
        image_reference=image_reference,
        oci_client=oci_client,
    )

    for layer in manifest.layers:
        blob = oci_client.blob(image_reference=image_reference, digest=layer.digest)

        with tarfile.open(
            fileobj=tarutil.FilelikeProxy(generator=blob.iter_content(chunk_size=4096)),
            mode='r|*',
        ) as tf:
            for info in tf:
                if long:
                    suffix = ''
                    if info.isdir():
                        prefix = 'd'
                    elif info.isfile():
                        prefix = 'f'
                        suffix = f'({info.size})'
                    elif info.issym():
                        suffix = f'-> {info.linkname}'
                        prefix = 's'
                    elif info.islnk():
                        prefix = 'l'
                    else:
                        prefix = ' '

                    print(f'{prefix} {info.name} {suffix}')
                else:
                    print(info.name)


def cat(
    image: str,
    path: str,
    outfile: str='-',
):
    if outfile == '-' and sys.stdout.isatty():
        print('error: either redirect output, or specify --outfile')
        exit(1)

    image_reference = oci.model.OciImageReference.to_image_ref(image)

    oci_client = ccc.oci.oci_client()
    manifest = _manifest(
        image_reference=image_reference,
        oci_client=oci_client,
    )

    for layer in manifest.layers:
        blob = oci_client.blob(image_reference=image_reference, digest=layer.digest)

        with tarfile.open(
            fileobj=tarutil.FilelikeProxy(generator=blob.iter_content(chunk_size=4096)),
            mode='r|*',
        ) as tf:
            for info in tf:
                if path.removeprefix('/') != info.name.removeprefix('/'):
                    continue
                break
            else:
                continue

            if not info.isfile():
                print(f'error: {path=} is not a regular file')
                exit(1)
            if outfile == '-':
                outfile = sys.stdout.buffer
            else:
                outfile = open(outfile, 'wb')

            octects_left = info.size
            while octects_left:
                read = min(octects_left, 4096)
                outfile.write(tf.fileobj.read(read))
                octects_left -= read

            # stop after first match
            exit(0)


def purge(image: str):
    oci_client = ccc.oci.oci_client()

    oci_client.delete_manifest(
        image_reference=image,
        purge=True,
    )
    print(f'purged {image}')


def purge_old(
    image: str,
    keep:int=128,
    skip_non_semver:bool=True,
):
    oci_client = ccc.oci.oci_client()
    pool = concurrent.futures.ThreadPoolExecutor(
        max_workers=8,
    )
    tags = [
        t for t in oci_client.tags(image_reference=image)
        if not skip_non_semver or version.is_semver_parseable(t)
    ]

    def sloppy_semver_parse(v: str):
        if v.count('-') > 1:
            v = v.split('-', 1)[0] + '-suffix' # discard other suffixes

        return version.parse_to_semver(v)

    tags = sorted([
        tag for tag in tags
        if not tag.startswith('latest')
        ],
      key=sloppy_semver_parse,
    ) # smallest version comes first

    remove_count = len(tags) - keep

    if remove_count < 1:
        print(f'will not remove images - there were less than {keep}')
        return

    print(f'found {len(tags)} image(s) - will purge {remove_count}')

    def purge_image(image_ref: str):
        print(f'purging {image_ref}')
        manifest = oci_client.manifest(
            image_reference=image_ref,
            accept=om.MimeTypes.prefer_multiarch,
        )
        try:
            oci_client.delete_manifest(
                image_reference=image_ref,
                purge=True,
                accept=om.MimeTypes.prefer_multiarch,
            )
        except requests.HTTPError as http_error:
            error_dict = http_error.response.json()
            errors = error_dict['errors']

            for e in errors:
                if e['code'] == 'GOOGLE_MANIFEST_DANGLING_PARENT_IMAGE':
                    msg = e['message']
                    parent_image_digest = msg.rsplit(' ', 1)[-1]
                    parent_img_ref = om.OciImageReference(image_ref)
                    print(f'warning: will purge dangling {parent_image_digest=}')
                    oci_client.delete_manifest(
                        image_reference=f'{parent_img_ref.ref_without_tag}@{parent_image_digest}',
                    )
            raise http_error

        if isinstance(manifest, om.OciImageManifest):
            return
        elif not isinstance(manifest, om.OciImageManifestList):
            raise ValueError(manifest)

        # manifest-list (aka multi-arch)
        image_ref = om.OciImageReference(image_ref)

        def iter_platform_refs():
            repository = image_ref.ref_without_tag
            base_tag = image_ref.tag

            for submanifest in manifest.manifests:
                p = submanifest.platform
                yield f'{repository}:{base_tag}-{p.os}-{p.architecture}'

        for ref in iter_platform_refs():
            if not oci_client.head_manifest(
                image_reference=ref,
                absent_ok=True,
            ):
                continue

            oci_client.delete_manifest(
                image_reference=ref,
                purge=True,
            )

    def iter_image_refs_to_purge():
        for idx, tag in enumerate(tags, 1):
            if idx > remove_count:
                print(f'stopping the purge at {tag}')
                return

            yield f'{image}:{tag}'

    for _ in pool.map(purge_image, iter_image_refs_to_purge()):
        pass


def _to_accept_mimetype(
    accept:OciManifestChoice=OciManifestChoice.PREFER_MULTIARCH,
):
    if accept is OciManifestChoice.SINGLE:
        accept = om.MimeTypes.single_image
    elif accept is OciManifestChoice.MULTIARCH:
        accept = om.MimeTypes.multiarch
    elif accept is OciManifestChoice.PREFER_MULTIARCH:
        accept = om.MimeTypes.prefer_multiarch
    else:
        raise NotImplementedError(accept)

    return accept


def manifest(
    image_reference: str,
    pretty:bool=True,
    accept:OciManifestChoice=OciManifestChoice.PREFER_MULTIARCH,
    print_expr=None,
):
    oci_client = ccc.oci.oci_client()

    accept = _to_accept_mimetype(accept)

    manifest_raw = oci_client.manifest_raw(
        image_reference=image_reference,
        accept=accept,
    )

    if pretty and not print_expr:
        manifest = oci.model.as_manifest(
            manifest=manifest_raw.text,
        )

        if isinstance(manifest, om.OciImageManifest):
            pprint.pprint(manifest.as_dict())
        else:
            pprint.pprint(manifest_raw.json())

        if isinstance(manifest, om.OciImageManifest):
            total_size = sum(blob.size for blob in manifest.blobs())
            manifest_raw_bytes = manifest_raw.content
            manifest_size = len(manifest_raw_bytes)
            manifest_digest = hashlib.sha256(manifest_raw_bytes).hexdigest()

            print()
            print(f'{total_size=} {manifest_digest=} {manifest_size=}')
        elif isinstance(manifest, om.OciImageManifestList):
            manifest_digest = hashlib.sha256(manifest_raw.content).hexdigest()
            print()
            print(f'{manifest_digest=}')

    elif not pretty and not print_expr:
        print(manifest_raw.text)
    else:
        manifest_bytes = manifest_raw.content
        manifest = oci.model.as_manifest(
            manifest=manifest_bytes,
        )

        # expose to eval
        size = len(manifest_bytes) # noqa
        digest = f'sha256:{hashlib.sha256(manifest_bytes).hexdigest()}' # noqa
        image_reference = oci.model.OciImageReference(image_reference) # noqa
        repository = image_reference.ref_without_tag # noqa

        print(eval(print_expr)) # nosec B307


def edit(
    image_reference: str,
    accept:OciManifestChoice=OciManifestChoice.PREFER_MULTIARCH,
    editor: str='vim',
):
    image_reference = om.OciImageReference(image_reference)
    oci_client = ccc.oci.oci_client()

    manifest_dict = oci_client.manifest_raw(
        image_reference=image_reference,
        accept=_to_accept_mimetype(accept),
    ).json()

    manifest_pretty = json.dumps(
        manifest_dict,
        indent=2,
    ).encode('utf-8')

    tmp = tempfile.NamedTemporaryFile()
    tmp.write(manifest_pretty)
    tmp.flush()

    # keep digest to find modifications
    digest = hashlib.sha256(manifest_pretty).hexdigest()

    res = subprocess.run((editor, tmp.name))
    if res.returncode != 0:
        print(f'editor returned {res.returncode=} - discarding changes!')
        exit(1)

    tmp.seek(0)
    file_digest = hashlib.sha256(tmp.read()).hexdigest()
    tmp.seek(0)

    if digest == file_digest:
        print('there were no changes - OCI Image Manifest will be left unchanged')
        exit(0)

    # upload altered manifest
    if image_reference.has_digest_tag:
        # if passed-in image-ref contained digest, target-ref must be patched
        image_reference = f'{image_reference.ref_without_tag}@sha256:{file_digest}'

    print(f'patching {image_reference}')
    oci_client.put_manifest(
        image_reference=image_reference,
        manifest=tmp.read(),
    )


def cfg(image_reference: str):
    oci_client = ccc.oci.oci_client()

    manifest = oci_client.manifest(image_reference=image_reference)
    cfg_blob = oci_client.blob(
        image_reference=image_reference,
        digest=manifest.config.digest,
        stream=False,
    ).content

    cfg_len = len(cfg_blob)
    cfg_dig = f'sha256:{hashlib.sha256(cfg_blob).hexdigest()}'

    pprint.pprint(json.loads(cfg_blob))

    print(f'{cfg_len} octetts')
    print(f'digest: {cfg_dig}')


def blob(
    image_reference: str,
    accept:OciManifestChoice=OciManifestChoice.SINGLE,
    digest: str=None,
    index:int=None,
    outfile: str='-',
):
    if not ((digest is not None) ^ (index is not None)):
        print('usage: exactly one of --digest, --index must be passed')
        exit(1)
    oci_client = ccc.oci.oci_client()

    if index is not None:
        manifest = oci_client.manifest(
            image_reference=image_reference,
            accept=_to_accept_mimetype(accept),
        )
        if not manifest.layers:
            print(f'{manifest.mediaType=}')
            print('image-manifest has no layers (hint: did you refer to a manifest-list?)')
            exit(1)
        try:
            digest = manifest.layers[index].digest
        except IndexError:
            print(f'no layer with {index=} present')
            exit(1)

    if outfile == '-':
        if sys.stdout.isatty():
            print('must not stream binary content to stdout (pipe to other process)')
            exit(1)
        outfh = sys.stdout
        write = outfh.buffer.write
    else:
        outfh = open(outfile, 'wb')
        write = outfh.write

    blob = oci_client.blob(
        image_reference=image_reference,
        digest=digest,
        stream=True,
    )
    for chunk in blob.iter_content(chunk_size=4096):
        write(chunk)

    outfh.flush()


def sanitise(image_reference: str):
    oci_client = ccc.oci.oci_client()

    manifest = oci_client.manifest(image_reference=image_reference)
    cfg_blob = oci_client.blob(
        image_reference=image_reference,
        digest=manifest.config.digest,
        stream=False,
    ).content

    if ow.is_cfg_blob_sane(manifest, cfg_blob=cfg_blob):
        print(f'{image_reference} was already sane - nothing to do')
        return

    patched_ref = ow.sanitise_image(image_ref=image_reference, oci_client=oci_client)

    print(patched_ref)


def osinfo(
    image_reference: str,
    delivery_cfg_name: str=None,
):
    oci_client = ccc.oci.oci_client()
    if not delivery_cfg_name:
        delivery_cfg_name = ctx.cfg.ctx.delivery_cfg_name
    if delivery_cfg_name:
        cfg_factory = ctx.cfg_factory()
        delivery_cfg = cfg_factory.delivery_endpoints(delivery_cfg_name)

        delivery_client = delivery.client.DeliveryServiceClient(
            routes=delivery.client.DeliveryServiceRoutes(
                base_url=delivery_cfg.base_url(),
            ),
        )
    else:
        delivery_client = None

    manifest = oci_client.manifest(
        image_reference=image_reference,
        accept=om.MimeTypes.prefer_multiarch,
    )
    if isinstance(manifest, om.OciImageManifestList):
        img_ref = om.OciImageReference(image_reference)
        sub_img_ref = f'{img_ref.ref_without_tag}@{manifest.manifests[0].digest}'

        manifest = oci_client.manifest(sub_img_ref)

    last_os_info = None

    for layer in manifest.layers:
        layer_blob = oci_client.blob(
            image_reference=image_reference,
            digest=layer.digest,
        )
        fileproxy = tarutil.FilelikeProxy(
            layer_blob.iter_content(chunk_size=tarfile.BLOCKSIZE)
        )
        tf = tarfile.open(fileobj=fileproxy, mode='r|*')
        if (os_info := us.determine_osinfo(tf)):
            last_os_info = os_info

    os_info = last_os_info
    pprint.pprint(os_info)

    if not delivery_client:
        print('no delivery-cfg found (use --delivery-cfg-name to configure)')
        print('will exit now')
        exit(0)

    os_infos = delivery_client.os_release_infos(
        os_id=os_info.ID,
        absent_ok=True,
    )

    if not os_infos:
        print(f'did not find os-infos for {os_info.ID=}')
        exit(0)

    branch_info = delivery.util.find_branch_info(
        os_id=os_info,
        os_infos=os_infos,
    )

    if not branch_info:
        print(f'did not find branch-info for {os_info.ID=} {os_info.VERSION=}')
        exit(1)

    print()
    print('Branch-Info:')
    print()
    pprint.pprint(branch_info)

    eol = delivery.util.branch_reached_eol(
        os_id=os_info,
        os_infos=os_infos,
    )

    have_update = delivery.util.update_available(
        os_id=os_info,
        os_infos=os_infos,
        ignore_if_patchlevel_is_next_to_greatest=False,
    )

    almost_up_to_date = not delivery.util.update_available(
        os_id=os_info,
        os_infos=os_infos,
        ignore_if_patchlevel_is_next_to_greatest=True,
    )

    distroless = os_info.is_distroless

    print(
        tabulate.tabulate(
            headers=('info', 'value'),
            tabular_data=(
                ('eol', eol,),
                ('update-available', have_update,),
                ('(almost)-up-to-date', almost_up_to_date,),
                ('distroless', distroless,),
            ),
        )
    )

    if have_update and almost_up_to_date:
        print()
        print('almost-up-to-date: not more than one patchlevel behind')



import ccc.oci
import container.util
import oci


def filter_image(
    source_ref:str,
    target_ref:str,
    remove_files:[str]=[],
):
    container.util.filter_image(
        source_ref=source_ref,
        target_ref=target_ref,
        remove_files=remove_files,
        mode=oci.ReplicationMode.PREFER_MULTIARCH,
    )


def to_digest_ref(image_ref:str):
    oci_client = ccc.oci.oci_client()
    print(oci_client.to_digest_hash(image_reference=image_ref))



__cmd_name__ = 'cosign'

import hashlib

import ccc.oci
import cosign
import signingserver
import oci.model as om


def _payload(
    image_reference: str,
):
    image_reference = om.OciImageReference(image_reference)
    oci_client = ccc.oci.oci_client()

    if not image_reference.has_tag:
        print('Error: image-ref must have tag')
        exit(1)
    if not image_reference.has_digest_tag:
        manifest = oci_client.manifest_raw(
            image_reference=image_reference,
            accept=om.MimeTypes.prefer_multiarch,
        )
        digest = hashlib.sha256(manifest.content).hexdigest()
        image_reference = f'{image_reference.ref_without_tag}@sha256:{digest}'

        print(f'resolved image-ref to {image_reference}')

    return cosign.payload_bytes(
        image_reference=image_reference,
    )


def signature_ref(
    image_reference: str,
):
    signature_image_ref = cosign.default_signature_image_reference(image_reference)
    print(signature_image_ref)


def sign(
    image_reference: str,
    signing_server_url: str,
    signing_server_client_cert: str,
    signing_server_client_cert_key: str,
    on_exist: cosign.OnExist=cosign.OnExist.APPEND,
    signing_server_certificate_ca: str=None,
    signing_algorithm: str='rsassa-pkcs1-v1_5',
    tls_validation: bool=True,
):
    image_reference = om.OciImageReference(image_reference)
    oci_client = ccc.oci.oci_client()

    if not image_reference.has_tag:
        print('Error: image-ref must have tag')
        exit(1)
    if not image_reference.has_digest_tag:
        manifest = oci_client.manifest_raw(
            image_reference=image_reference,
            accept=om.MimeTypes.prefer_multiarch,
        )
        digest = hashlib.sha256(manifest.content).hexdigest()
        image_reference = f'{image_reference.ref_without_tag}@sha256:{digest}'

        print(f'resolved image-ref to {image_reference}')

    payload_bytes = _payload(
        image_reference=image_reference,
    )

    signingserver_client = signingserver.SigningserverClient(
        cfg=signingserver.SigningserverClientCfg(
            base_url=signing_server_url,
            client_certificate=signing_server_client_cert,
            client_certificate_key=signing_server_client_cert_key,
            server_certificate_ca=signing_server_certificate_ca,
            validate_tls_certificate=tls_validation,
        ),
    )

    signature = signingserver_client.sign(
        content=payload_bytes,
        signing_algorithm=signing_algorithm,
    )

    signature_image_ref = cosign.default_signature_image_reference(image_reference)
    print(f'creating signature: {signature_image_ref}')

    cosign.sign_image(
        image_reference=image_reference,
        signature_image_reference=signature_image_ref,
        signature=signature.signature.replace('\n', ''),
        signing_algorithm=signing_algorithm,
        public_key=signature.public_key.replace('\n', ''),
        on_exist=on_exist,
        oci_client=oci_client,
        payload=payload_bytes,
    )



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import argparse
import enum
import functools
import inspect
import itertools
import os
import pkgutil
import sys
import types

try:
    import ci.util
except ModuleNotFoundError:
    repo_dir = os.path.abspath(
        os.path.join(
            os.path.dirname(__name__),
            os.pardir,
            os.pardir,
        )
    )
    sys.path.insert(1, repo_dir)
    import ci.util

import ci.log
# to overwrite cli.py log level, call
# "configure_default_logging(force=True, stdout_level=logging.DEBUG)" in specific module cli
ci.log.configure_default_logging(force=True)

import ctx  # noqa: E402


cfg = ctx.cfg
terminal_cfg = cfg.terminal

if terminal_cfg and terminal_cfg.output_columns is not None:
    column_width = terminal_cfg.output_columns
    # Create a custom width formatter by fixing two arguments for the default formatter class,
    # namely 'width' (defaults to 80 - 2) and 'max_help_position' (defaults to 24)
    FORMATTER_CLASS = functools.partial(
        argparse.RawDescriptionHelpFormatter,
        max_help_position=24,
        width=column_width
    )
else:
    FORMATTER_CLASS = argparse.RawDescriptionHelpFormatter


def _subcommand_modules() -> dict[str, str]:
    '''
    returns a dict where keys are subcommand-names, and values are module-names
    (module-names are sometimes prefixed with an underscore in this directory to avoid
    name conflicts)
    '''
    fnames = [
        n for n in os.listdir(os.path.dirname(os.path.realpath(__file__)))
        if n.endswith('.py') and not n == os.path.basename(__file__) and not n == '__init__.py'
    ]

    return {
        n.removeprefix('_').removesuffix('.py'): n.removesuffix('.py')
        for n in fnames
    }


def main():
    '''
    Creates a command line parser (using argparse) for each python module found in this
    directory (except for _this_ module). For each module, a sub-command named as the
    module name is added. Each function defined in a given module is again added as a
    sub-sub-command. Based on the function signature, optional arguments are added.
    This parser is then used to parse the given ARGV. Provided that parsing succeeds,
    the thus specified function is executed.
    '''
    subcommand_modules = _subcommand_modules()
    # as an optimisation, try to parse chosen subcommand (if any). If there is one, we can skip
    # loading any other modules (which saves a lot of startup time).
    parser = argparse.ArgumentParser()
    parser.add_argument('subcommand', nargs=1, choices=subcommand_modules.keys())
    add_global_args(parser)
    parsed, _ = parser.parse_known_args()
    if parsed.subcommand:
        subcommand = parsed.subcommand[0]
        chosen_module_name = subcommand_modules[subcommand]
    else:
        subcommand = None
        chosen_module_name = None

    parser = argparse.ArgumentParser(formatter_class=FORMATTER_CLASS)
    add_global_args(parser)
    sub_command_parsers = parser.add_subparsers()
    cli_module_dir = os.path.dirname(os.path.abspath(os.path.realpath(__file__)))
    sys.path.insert(0, cli_module_dir)
    for _, module_name, _ in pkgutil.iter_modules([cli_module_dir]):
        # if subcommand was given, only load module for chosen subcommand
        if chosen_module_name and module_name != chosen_module_name:
            continue
        # skip own module name
        if module_name == os.path.splitext(os.path.basename(__file__))[0]:
            continue
        if module_name == 'setup': # skip setup.py
            continue
        add_module(module_name, sub_command_parsers)
    if len(sys.argv) == 1:
        parser.print_usage()
        sys.exit(1)
    parsed = parser.parse_args()
    # write parsed args to global ctx module so called module functions may
    # retrieve if (see ci.util.ctx)
    ctx.args = parsed

    ctx.load_config()

    # mark 'cli' mode
    ci.util._set_cli(True)
    if hasattr(parsed, 'module'):
        parsed.module.args = parsed
        parsed.func(parsed)


def add_global_args(parser):
    parser.add_argument('--quiet', action='store_true')
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('--cfg-dir', default=None)


def add_module(module_name, parser):
    module = __import__(module_name)

    # skip if module defines a symbol 'main'
    if hasattr(module, 'main'):
        return
    if hasattr(module, '__cmd_name__'):
        cmd_name = module.__cmd_name__
    else:
        cmd_name = module_name

    module_parser = parser.add_parser(cmd_name, formatter_class=FORMATTER_CLASS)
    module_parser.set_defaults(
      func=display_usage_function(module_parser),
      module=module
    )
    # add module-specific arguments
    if hasattr(module, '__add_module_command_args'):
        getattr(module, '__add_module_command_args')(module_parser)

    function_parsers = module_parser.add_subparsers()

    for fname, function in inspect.getmembers(module, predicate=inspect.isfunction):
        if fname.startswith('_'):
            continue # skip "private" functions
        function_docstring = inspect.getdoc(function)
        function_parser = function_parsers.add_parser(
            fname,
            description=function_docstring,
            formatter_class=FORMATTER_CLASS,
        )
        fspec = inspect.getfullargspec(function)
        function_parser.set_defaults(func=run_function(function))

        action = None
        # defaults are filled "from the end", so reverse both argnames and defaults
        for argname, default in reversed(list(
            itertools.zip_longest(
              reversed(fspec.args),
              reversed(fspec.defaults or []),
              fillvalue=NotImplemented # workaround to be able to discriminate from None
            )
          )):
            cl_arg = '--' + argname.replace('_', '-')
            annotation = fspec.annotations.get(argname, None)
            argtype = None
            action = None
            kwargs = {}
            if annotation:
                from ci.util import CliHint
                # special case: CliHint
                if type(annotation) == CliHint:
                    typehint = annotation.typehint
                    kwargs.update(annotation.argparse_args)
                else:
                    typehint = annotation
                # handle type-specific actions (lists, booleans, ..)
                if type(typehint) == type: # primitives (str, bool, int, ..)
                    argtype = typehint
                    if typehint == bool:
                        action = 'store_true'
                        argtype = None # type must not be set for store_true/store_false actions
                elif type(typehint) == list:
                    action = 'append'
                elif isinstance(typehint, types.GenericAlias):
                    # assume it is a list/sequence
                    type_args = typehint.__args__
                    if len(type_args) == 1:
                        argtype = type_args[0]
                        action = 'append'
                    else:
                        print('Error: GenericAlias must have exactly one type-parameter')
                        exit(1)

                elif callable(typehint):
                    argtype = typehint
                    if inspect.isclass(typehint) and issubclass(typehint, enum.Enum):
                        # XXX: improve online-help
                        kwargs['choices'] = [
                            e for e in typehint
                        ]

            if default != NotImplemented:
                required = False
            else:
                required = True
                default = None # set back to None to not have argparser behave strangely :-)

            # add_argument does not allow 'type' as a parameter in some cases;
            # workaround this by omitting it in all cases where it is None anyway
            if argtype is not None and 'type' not in kwargs:
                kwargs['type'] = argtype

            if action:
                kwargs['action'] = action

            if default:
                help_text = kwargs.get('help', '')
                help_text += '(default: %(default)s)'
                kwargs['help'] = help_text

            function_parser.add_argument(
              cl_arg,
              required=required,
              default=default,
              **kwargs
            )

            if annotation == bool and not argname.startswith('no'):
                kwargs['help'] = '(default: False)'
                cl_arg = '--no-' + argname.replace('_', '-')
                kwargs['action'] = 'store_false'
                function_parser.add_argument(
                  cl_arg,
                  required=False,
                  dest=argname.replace('-', '_'),
                  **kwargs
                )


def run_function(function):
    def function_runner(args):
        fspec = inspect.getfullargspec(function)
        function_args = []
        for argname in fspec.args:
            function_args.append(getattr(args, argname))
        function(*function_args)
    return function_runner


def display_usage_function(parser):
    def display_usage(_):
        parser.print_usage()
    return display_usage


if __name__ == '__main__':
    main()



import signingserver


def sign(
    filepath: str,
    signing_server_url: str,
    signing_server_client_cert: str,
    signing_server_client_cert_key: str,
    output: str='signature',
    signing_server_certificate_ca: str=None,
    signing_algorithm: str='rsassa-pkcs1-v1_5',
    tls_validation: bool=True,
):
    with open(filepath, 'rb') as f:
        payload_bytes = f.read()

    signingserver_client = signingserver.SigningserverClient(
        cfg=signingserver.SigningserverClientCfg(
            base_url=signing_server_url,
            client_certificate=signing_server_client_cert,
            client_certificate_key=signing_server_client_cert_key,
            server_certificate_ca=signing_server_certificate_ca,
            validate_tls_certificate=tls_validation,
        ),
    )

    signature = signingserver_client.sign(
        content=payload_bytes,
        signing_algorithm=signing_algorithm,
    )

    if output == 'signature':
        print(signature.signature.replace('\n', ''), end='')
    elif output == 'raw':
        print(signature.raw)
    elif output == 'certificate':
        print(signature.certificate)
    elif output == 'public-key':
        print(signature.public_key)
    else:
        print(f'invalid choice: {output=}')
        exit(1)



import setuptools
import shutil
import os

own_dir = os.path.abspath(os.path.dirname(__file__))


def requirements():
    yield f'gardener-cicd-libs=={version()}'


def modules():
    return []
    return [
        os.path.basename(os.path.splitext(module)[0]) for module in
        os.scandir(path=own_dir)
        if module.is_file() and module.name.endswith('.py')
    ]


def version():
    with open(os.path.join(own_dir, 'gardener_ci', 'VERSION')) as f:
        return f.read().strip()


# cp scripts
src_bin_dir = os.path.join(own_dir, os.pardir, 'bin')
tgt_bin_dir = os.path.join(own_dir, 'bin')
shutil.copytree(src_bin_dir, tgt_bin_dir)


setuptools.setup(
    name='gardener-cicd-cli',
    version=version(),
    description='Gardener CI/CD Command Line Interface',
    python_requires='>=3.11',
    py_modules=modules(),
    packages=setuptools.find_packages(),
    package_data={
        'gardener_ci': ['VERSION'],
    },
    install_requires=list(requirements()),
    scripts=[os.path.join(tgt_bin_dir, 'purge_history')],
    entry_points={
        'console_scripts': [
            'gardener-ci = gardener_ci.cli_gen:main',
            'cli.py = gardener_ci.cli_gen:main', # XXX backwards-compatibilty - rm this
            'yaml2json = gardener_ci.yaml2json:main'
        ],
    },
)



import json
import os

import model.container_registry as mcr
import oci.auth as oa


def cfg(
    image_ref_prefixes: [str],
    privileges: str ='readwrite',
    patch: bool=False,
):
    cfgs = set()
    if privileges == 'readwrite':
        privileges = oa.Privileges.READWRITE
    elif privileges in ('readonly', 'read'):
        privileges = oa.Privileges.READONLY
    else:
        raise ValueError(privileges)

    for prefix in image_ref_prefixes:
        cfg = mcr.find_config(
            image_reference=prefix,
            privileges=privileges,
        )
        if not cfg:
            continue

        cfgs.add(cfg)

    if not cfgs:
        print(f'did not find any cfg for given prefixes {image_ref_prefixes=}')
        exit(1)

    def iter_auths():
        for cfg in cfgs:
            for host, auth in cfg.as_docker_auths().items():
                yield host, auth

    docker_cfg = {
        'auths': dict(iter_auths()),
    }

    if not patch:
        print(json.dumps(docker_cfg, indent=2))
        exit(0)

    with open(os.path.join(os.environ['HOME'], '.docker', 'config.json'), 'w') as f:
        json.dump(docker_cfg, f)



import concurrent.futures
import logging
import sys

import gci.componentmodel as cm

import ccc.oci
import cnudie.iter
import cnudie.purge
import cnudie.retrieve
import cnudie.util
import cnudie.validate
import ctx
import version

logger = logging.getLogger(__name__)


def _ocm_lookup(ocm_repo: str=None):
    if ocm_repo:
        return cnudie.retrieve.create_default_component_descriptor_lookup(
            ocm_repository_lookup=cnudie.retrieve.ocm_repository_lookup(ocm_repo),
        )
    else:
        return ctx.cfg.ctx.ocm_lookup


def retrieve(
    name: str,
    version: str=None,
    ocm_repo: str=None,
    out: str=None
):
    if not version:
        name, version = name.rsplit(':', 1)

    ocm_lookup = _ocm_lookup(ocm_repo=ocm_repo)

    component_descriptor = ocm_lookup(
        cm.ComponentIdentity(
            name=name,
            version=version,
        ),
    )

    if not component_descriptor:
        print(f'Error: did not find {name}:{version}')
        exit(1)

    if out:
        outfh = open(out, 'w')
    else:
        outfh = sys.stdout

    component_descriptor.to_fobj(fileobj=outfh)
    outfh.flush()
    outfh.close()


def validate(
    name: str,
    version: str,
    ocm_repo: str=None,
    out: str=None
):
    ocm_lookup = _ocm_lookup(ocm_repo=ocm_repo)

    logger.info('retrieving component-descriptor..')
    component_descriptor = ocm_lookup(
        cm.ComponentIdentity(
            name=name,
            version=version,
        ),
    )
    component = component_descriptor.component
    logger.info('validating component-descriptor..')

    violations = tuple(
        cnudie.validate.iter_violations(
            nodes=cnudie.iter.iter(
                component=component,
                recursion_depth=0,
            ),
        )
    )

    if not violations:
        logger.info('component-descriptor looks good')
        return

    logger.warning('component-descriptor yielded validation-errors (see below)')
    print()

    for violation in violations:
        print(violation.as_error_message)


def ls(
    name: str,
    greatest: bool=False,
    final: bool=False,
    ocm_repo: str=None,
):
    if ocm_repo:
        ocm_repo_lookup = cnudie.retrieve.ocm_repository_lookup(ocm_repo)
    else:
        ocm_repo_lookup = ctx.cfg.ctx.ocm_repository_lookup

    version_lookup = cnudie.retrieve.version_lookup(ocm_repository_lookup=ocm_repo_lookup)

    ocm_repo = next(ocm_repo_lookup(name))

    if greatest:
        print(version.greatest_version(
            versions=version_lookup(name),
        ))
        return

    versions = version_lookup(name)

    for v in versions:
        if final:
            parsed_version = version.parse_to_semver(v)
            if parsed_version.prerelease:
                continue
        print(v)


def purge_old(
    name: str,
    final: bool=False,
    ocm_repo: str=None,
    keep: int=256,
    threads: int=32,
):
    if ocm_repo:
        ocm_repo_lookup = cnudie.retrieve.ocm_repository_lookup(ocm_repo)
    else:
        ocm_repo_lookup = ctx.cfg.ctx.ocm_repository_lookup

    ocm_repo = next(ocm_repo_lookup(name))

    version_lookup = cnudie.retrieve.version_lookup(ocm_repository_lookup=ocm_repo_lookup)

    versions = version_lookup(name)

    if not final:
        versions = [
            v for v in versions
            if not version.parse_to_semver(v).prerelease
        ]

    versions = version.smallest_versions(
        versions=versions,
        keep=keep,
    )

    print(f'will rm {len(versions)} version(s) using {threads=}')

    pool = concurrent.futures.ThreadPoolExecutor(max_workers=threads)
    oci_client = ccc.oci.oci_client(
        http_connection_pool_size=threads,
    )

    def purge_component_descriptor(ref: str):
        oci_client.delete_manifest(
            image_reference=ref,
            purge=True,
        )
        print(f'purged: {ref}')

    def iter_oci_refs_to_rm():
        for v in versions:
            ref = f'{ocm_repo}/component-descriptors/{name}:{v}'
            yield pool.submit(
                purge_component_descriptor,
                ref=ref,
            )

    for ref in concurrent.futures.as_completed(iter_oci_refs_to_rm()):
        pass


def purge(
    name: str,
    recursive: bool=False,
    version: str=None,
    ocm_repo: str=None,
):
    if not version:
        name, version = name.rsplit(':', 1)

    ocm_lookup = _ocm_lookup(ocm_repo=ocm_repo)

    lookup = cnudie.retrieve.oci_component_descriptor_lookup()

    component_descriptor = ocm_lookup(
        cm.ComponentIdentity(
            name=name,
            version=version,
        ),
    )

    oci_client = ccc.oci.oci_client()

    cnudie.purge.remove_component_descriptor_and_referenced_artefacts(
        component=component_descriptor.component,
        oci_client=oci_client,
        lookup=lookup,
        recursive=recursive,
    )
